% [TRAJNEW, BESTIDX] = FASTTRELLIS(Z, X, PARAMS, VERBOSITY)
% Inputs: 
% X			noisy piecewise linear data 
% PARAMS		structure of input parameters
% 			PARAMS.M = number of quantization levels
%			PARAMS.N = length of reconstructed vector
%			PARAMS.SIG = standard deviation of gaussian noisy
%			PARAMS.P = probability of staying in same slope value
% VERBOSITY		sets verbosity 1=verbose, 0=quiet (default is 1)
%
% Outputs:
% TRAJNEW 		holds the optimal trajectories coming into each node at
%			the maximum trellis depth, and the associated likelihood
%			value in the last column
% BESTIDX		index of the row in TRAJNEW with largest likelihood
%
% Remarks:
% 1) Data with slopes in [0,1] should be rescaled by a factor of M-1, this allows
% the trellis search to run on integers {0, 1, ..., M-1} and avoid numerical issues
% with small fractional values. Input noise standard deviation should also be scaled
% by (M-1) to maintain consistency.
% 2) Typical usage of output involves picking out the MAP optimal sequence of
% slope values as DIFF(TRAJNEW(BESTIDX, 1:END-1)).
% The associated likelihood value is stored in TRAJNEW(BESTIDX, END).
% In rare situations there may be ties, causing BESTIDX to contain more than 1 member.
% 3) There are slight differences from the presentation in the paper because MATLAB 
% indexing begins at 1 instead of 0.
% Ver 2, June 1, 2012

function [trajnew, bestidx] = fasttrellis(x, params, varargin)

M = params.M;
N = params.N;
sig =params.sig;
d = 1; mu = 0;	% Unused variables. Reserved for future versions.
p = params.p;

if isempty(varargin) || varargin{1}==1
    disp('I will be verbose here.');
    verbose = 1;
else
    verbose = 0;
end

trajold = zeros(M,3);
for i=1:M
    % we are at depth 2 now
    trajold(i,:) = [0, d*(i-1), ...
        loglike(x(2), d*(i-1), d*(i-1), 0, mu, sig, p, M)]; 
end

for depth = 3:N
    if verbose, disp(['At depth ', num2str(depth), '/', num2str(N)]); end
          
    trajnew = zeros((depth-1)*(M-1)+1, depth+1);
    if depth==N, trajnewtracker = trajnew; end
    
    for nodeNum = 1:(depth-1)*(M-1)+1
        if nodeNum < M
            prevNodeNum = 1:nodeNum;
        elseif nodeNum > (depth-2)*(M-1)+1
            prevNodeNum = nodeNum-M+1:(depth-2)*(M-1)+1;
        else
            prevNodeNum = nodeNum-M+1:nodeNum;
        end
        
        maxearned = -Inf;
        for j = 1:length(prevNodeNum)
            earning = trajold(prevNodeNum(j),end) + ...
                loglike(  x(depth),...
                d*(nodeNum-1), ...
                d*(nodeNum-prevNodeNum(j)), ...
                (trajold(prevNodeNum(j),end-1) ...
                    - trajold(prevNodeNum(j), end-2)), mu, sig, p, M  );
            
            if earning >= maxearned
                trajnew(nodeNum, :) = ...
                    [trajold(prevNodeNum(j),1:end-1), ...
                    d*(nodeNum-1), earning]; 
                if depth==N, trajnewtracker(nodeNum,:) =...
                        ones(size(trajnewtracker(nodeNum,:))); end
                maxearned = earning;
            end
        end
    end
    trajold = trajnew;
    %plot( trajnew(:,end) ); title(num2str(numel(trajnew(:,end)))); 
    %drawnow; pause(0.1);
end

bestidx = find(trajnew(:,end) == max(trajnew(:,end)));

end

function val = loglike(x, z, snew, sold, mu, sig, p, M)
	%val = log(Q(x-z, mu, sig)) + log(P(snew, sold, p, M));
    %val = logQ(x-z, mu, sig) + log(P(snew, sold, p, M));
    val = -1/2*log(2*pi*sig^2) - ((x-z)-mu)^2 / 2 / sig^2 + ...
        log(p*(snew==sold) + (1-p)/(M-1)*(1-(snew==sold)));
end

function t1 = logQ(v, mu, sig)
    t1 = -1/2*log(2*pi*sig^2) - (v-mu)^2 / 2 / sig^2;
end

function t1 = Q(v, mu, sig)
    %t1 = pdf('Normal', v, mu, sig);
    t1 = 1/sqrt(2*pi*sig^2)*exp( -(v-mu)^2 / 2 / sig^2 );
end

function t2 = P(s2, s1, p, M)
    t2 = p*(s2==s1) + (1-p)/(M-1)*(1-(s2==s1));
end
