% FUNCTION [ZEST, SIG2CUR, PPCUR] = MAPSLOPE(X, M, SIG2OLD, PPOLD, THRESH, MAXIT)
% Inputs:
% X			Noisy input data vector (slopes must be in [0,1])
% M			Number of quantized slope values
% sig2old		Starting guess of noise variance \sigma^2
% ppold			Starting guess of p=probability of maintaining slope constant
% thresh		Termination criterion for change in likelihood function values
% maxit			Maximum number of iteration before terminating
%
% Outputs:
% ZEST		estimated piecewise linear function values, take derivative for slope values
% sig2cur	estimated noise variance
% ppcur		estimated value of prob. of staying in same slope value
%
% Remarks:
% Ensure your data has slope values in the range [0,1]. 
% This function rescales the data and the noise standard deviation by a 
% factor of (M-1) when calling FASTTRELLIS. (This rescaling trick allows 
% the routine to search over the integer values {0, 1, ..., M-1}).
% 
% v.1 Jan 31, 2013
% v.2 Oct 06, 2013 fixed rescaling by (M-1) so top level script doesn't
%                  have to worry about it any more

function [zest, sig2cur, ppcur] = mapslope(x, M, sig2old, ppold, thresh, maxit)

% Initialize guesses for Viterbi+Restoration Maximization
ppcur = ppold;
sig2cur = sig2old;
Lold = -Inf;
%sig2old = 0.6; sig2cur = 0.6;
disp(['sig2old = ', num2str(sig2old), ', ppold = ', num2str(ppold)]);
N = numel(x);
verbosity = 0;

% Iterate 
iterated = 0;
while(1)
	disp(['At iteration: ', num2str(iterated+1)]);
	
	params = struct('N', N, 'M', M, 'sig', (M-1)*sqrt(sig2old), 'p', ppold);
	 
	[t, b] = fasttrellis((M-1)*x, params, verbosity);
	zest = t(b,1:end-1); 
	
	ppcur = 1-sum(abs(diff(diff(zest))) > 0.0001)/N;
    zest = zest/(M-1); % Rescale (see HELP in mapslope.m)
	sig2cur = var(x(:) - zest(:));
    Lcur = t(b, end); 
    
	ppold = ppcur;
	sig2old = sig2cur;
	
	disp(['sig2cur = ', num2str(sig2cur), ', ppcur = ', num2str(ppcur)]);
	disp(['|Lcur-Lold|=',num2str(abs(Lcur-Lold))]);
	iterated = iterated+1;
	if iterated>maxit || abs(Lold-Lcur)<thresh
    	break;
	end
	Lold = Lcur;
	disp(' ');
end
